@@ -27,7 +27,7 @@ from concurrent.futures import ThreadPoolExecutor
 
 from .exceptions import FailedRequestError, InvalidRequestError
 
-VERSION = '1.1.6'
+VERSION = '1.1.10'
 
 
 class HTTP:
@@ -68,6 +68,10 @@ class HTTP:
     :param max_retries: The number of times to re-attempt a request.
     :type max_retries: int
 
+    :param retry_delay: Seconds between retries for returned error or timed-out
+        requests. Default is 3 seconds.
+    :type retry_delay: int
+
     :param referral_id: An optional referer ID can be added to each request for
         identification.
     :type referral_id: str
@@ -79,7 +83,7 @@ class HTTP:
     def __init__(self, endpoint, api_key=None, api_secret=None,
                  logging_level=logging.INFO, log_requests=False,
                  request_timeout=10, recv_window=5000, force_retry=False,
-                 max_retries=3, referral_id=None):
+                 max_retries=3, retry_delay=3, referral_id=None):
         """Initializes the HTTP class."""
 
         # Set the endpoint.
@@ -104,6 +108,10 @@ class HTTP:
         self.recv_window = recv_window
         self.force_retry = force_retry
         self.max_retries = max_retries
+        self.retry_delay = retry_delay
+
+        # Set whitelist of non-fatal Bybit return error codes to retry
+        self.retry_codes = {10002, 10006, 30034, 30035, 130035, 130150}
 
         # Initialize requests session.
         self.client = requests.Session()
@@ -148,6 +156,11 @@ class HTTP:
         :returns: Request results as dictionary.
         """
 
+        # Replace query param 'from_time' since 'from' keyword is reserved.
+        # Temporary workaround until Bybit updates official request params
+        if 'from_time' in kwargs:
+            kwargs['from'] = kwargs.pop('from_time')
+
         if 'symbol' in kwargs and 'USDT' in kwargs['symbol']:
             suffix = '/public/linear/kline'
         else:
@@ -184,6 +197,11 @@ class HTTP:
         :returns: Request results as dictionary.
         """
 
+        # Replace query param 'from_id' since 'from' keyword is reserved.
+        # Temporary workaround until Bybit updates official request params
+        if 'from_id' in kwargs:
+            kwargs['from'] = kwargs.pop('from_id')
+
         if 'symbol' in kwargs and 'USDT' in kwargs['symbol']:
             suffix = '/public/linear/recent-trading-records'
         else:
@@ -217,6 +235,11 @@ class HTTP:
         :returns: Request results as dictionary.
         """
 
+        # Replace query param 'from_id' since 'from' keyword is reserved.
+        # Temporary workaround until Bybit updates official request params
+        if 'from_id' in kwargs:
+            kwargs['from'] = kwargs.pop('from_id')
+
         return self._submit_request(
             method='GET',
             path=self.endpoint + '/v2/public/liq-records',
@@ -232,6 +255,11 @@ class HTTP:
         :returns: Request results as dictionary.
         """
 
+        # Replace query param 'from_time' since 'from' keyword is reserved.
+        # Temporary workaround until Bybit updates official request params
+        if 'from_time' in kwargs:
+            kwargs['from'] = kwargs.pop('from_time')
+
         if 'symbol' in kwargs and 'USDT' in kwargs['symbol']:
             suffix = '/public/linear/mark-price-kline'
         else:
@@ -346,7 +374,7 @@ class HTTP:
         if 'symbol' in kwargs and 'USDT' in kwargs['symbol']:
             suffix = '/private/linear/order/list'
         else:
-            suffix = '/open-api/order/list'
+            suffix = '/v2/private/order/list'
 
         return self._submit_request(
             method='GET',
@@ -433,7 +461,7 @@ class HTTP:
         if 'symbol' in kwargs and 'USDT' in kwargs['symbol']:
             suffix = '/private/linear/order/replace'
         else:
-            suffix = '/open-api/order/replace'
+            suffix = '/v2/private/order/replace'
 
         return self._submit_request(
             method='POST',
@@ -498,7 +526,7 @@ class HTTP:
         if 'symbol' in kwargs and 'USDT' in kwargs['symbol']:
             suffix = '/private/linear/stop-order/create'
         else:
-            suffix = '/open-api/stop-order/create'
+            suffix = '/v2/private/stop-order/create'
 
         return self._submit_request(
             method='POST',
@@ -542,7 +570,7 @@ class HTTP:
         if 'symbol' in kwargs and 'USDT' in kwargs['symbol']:
             suffix = '/private/linear/stop-order/list'
         else:
-            suffix = '/open-api/stop-order/list'
+            suffix = '/v2/private/stop-order/list'
 
         return self._submit_request(
             method='GET',
@@ -564,7 +592,7 @@ class HTTP:
         if 'symbol' in kwargs and 'USDT' in kwargs['symbol']:
             suffix = '/private/linear/stop-order/cancel'
         else:
-            suffix = '/open-api/stop-order/cancel'
+            suffix = '/v2/private/stop-order/cancel'
 
         return self._submit_request(
             method='POST',
@@ -629,7 +657,7 @@ class HTTP:
         if 'symbol' in kwargs and 'USDT' in kwargs['symbol']:
             suffix = '/private/linear/stop-order/replace'
         else:
-            suffix = '/open-api/stop-order/replace'
+            suffix = '/v2/private/stop-order/replace'
 
         return self._submit_request(
             method='POST',
@@ -721,16 +749,21 @@ class HTTP:
 
     def set_leverage(self, **kwargs):
         """
-        For linear markets only. Change user leverage.
+        Change user leverage.
 
         :param kwargs: See
             https://bybit-exchange.github.io/docs/linear/#t-setleverage.
         :returns: Request results as dictionary.
         """
 
+        if 'symbol' in kwargs and 'USDT' in kwargs['symbol']:
+            suffix = '/private/linear/position/set-leverage'
+        else:
+            suffix = '/v2/private/position/leverage/save'
+
         return self._submit_request(
             method='POST',
-            path=self.endpoint + '/private/linear/position/set-leverage',
+            path=self.endpoint + suffix,
             query=kwargs,
             auth=True
         )
@@ -763,7 +796,7 @@ class HTTP:
 
         return self._submit_request(
             method='POST',
-            path=self.endpoint + '/position/change-position-margin',
+            path=self.endpoint + '/v2/private/position/change-position-margin',
             query=kwargs,
             auth=True
         )
@@ -780,7 +813,7 @@ class HTTP:
         if 'symbol' in kwargs and 'USDT' in kwargs['symbol']:
             suffix = '/private/linear/position/trading-stop'
         else:
-            suffix = '/open-api/position/trading-stop'
+            suffix = '/v2/private/position/trading-stop'
 
         return self._submit_request(
             method='POST',
@@ -807,16 +840,19 @@ class HTTP:
 
     def user_leverage(self, **kwargs):
         """
-        Get user leverage.
+        ABANDONED! Please use my_position instead. Fetches user leverage by
+        fetching user position.
 
         :param kwargs: See
             https://bybit-exchange.github.io/docs/inverse/#t-getleverage.
         :returns: Request results as dictionary.
         """
 
+        self.logger.warn('This endpoint is deprecated and will be removed. Use my_position()')
+
         return self._submit_request(
             method='GET',
-            path=self.endpoint + '/user/leverage',
+            path=self.endpoint + '/v2/private/position/list',
             query=kwargs,
             auth=True
         )
@@ -830,6 +866,8 @@ class HTTP:
         :returns: Request results as dictionary.
         """
 
+        self.logger.warn('This endpoint is deprecated and will be removed. Use set_leverage()')
+
         return self._submit_request(
             method='POST',
             path=self.endpoint + '/user/leverage/save',
@@ -931,7 +969,7 @@ class HTTP:
         if 'symbol' in kwargs and 'USDT' in kwargs['symbol']:
             suffix = '/public/linear/funding/prev-funding-rate'
         else:
-            suffix = '/open-api/funding/prev-funding-rate'
+            suffix = '/v2/private/funding/prev-funding-rate'
 
         return self._submit_request(
             method='GET',
@@ -955,7 +993,7 @@ class HTTP:
         if 'symbol' in kwargs and 'USDT' in kwargs['symbol']:
             suffix = '/private/linear/funding/prev-funding'
         else:
-            suffix = '/open-api/funding/prev-funding'
+            suffix = '/v2/private/funding/prev-funding'
 
         return self._submit_request(
             method='GET',
@@ -976,7 +1014,7 @@ class HTTP:
         if 'symbol' in kwargs and 'USDT' in kwargs['symbol']:
             suffix = '/private/linear/funding/predicted-funding'
         else:
-            suffix = '/open-api/funding/predicted-funding'
+            suffix = '/v2/private/funding/predicted-funding'
 
         return self._submit_request(
             method='GET',
@@ -994,7 +1032,7 @@ class HTTP:
 
         return self._submit_request(
             method='GET',
-            path=self.endpoint + '/open-api/api-key',
+            path=self.endpoint + '/v2/private/account/api-key',
             auth=True
         )
 
@@ -1044,9 +1082,14 @@ class HTTP:
         :returns: Request results as dictionary.
         """
 
+        # Replace query param 'from_id' since 'from' keyword is reserved.
+        # Temporary workaround until Bybit updates official request params
+        if 'from_id' in kwargs:
+            kwargs['from'] = kwargs.pop('from_id')
+
         return self._submit_request(
             method='GET',
-            path=self.endpoint + '/open-api/wallet/fund/records',
+            path=self.endpoint + '/v2/private/wallet/fund/records',
             query=kwargs,
             auth=True
         )
@@ -1062,7 +1105,7 @@ class HTTP:
 
         return self._submit_request(
             method='GET',
-            path=self.endpoint + '/open-api/wallet/withdraw/list',
+            path=self.endpoint + '/v2/private/wallet/withdraw/list',
             query=kwargs,
             auth=True
         )
@@ -1126,43 +1169,31 @@ class HTTP:
         """
 
         # First we fetch the user's position.
-        r = self._submit_request(
-            method='GET',
-            path=self.endpoint + '/v2/private/position/list',
-            query={'symbol': symbol},
-            auth=True
-        )
-
-        # Next we detect the position size and side.
         try:
-            position = r['result']['size']
-            side = 'Buy' if r['result']['side'] == 'Sell' else 'Sell'
-            if position == 0:
-                return self.logger.error('No position detected.')
+            r = self.my_position(symbol=symbol)['result']
 
         # If there is no returned position, we want to handle that.
-        except IndexError:
+        except KeyError:
             return self.logger.error('No position detected.')
 
-        # We should know if we're pinging the inverse or linear perp API.
-        if symbol.endswith('USDT'):
-            close_path = self.endpoint + '/private/linear/order/create'
-        else:
-            close_path = self.endpoint + '/v2/private/order/create'
-
-        # Submit a market order against your position for the same qty.
-        return self._submit_request(
-            method='POST',
-            path=close_path,
-            query={
+        # Next we generate a list of market orders
+        orders = [
+            {
                 'symbol': symbol,
                 'order_type': 'Market',
-                'side': side,
-                'qty': position,
-                'time_in_force': 'ImmediateOrCancel'
-            },
-            auth=True
-        )
+                'side': 'Buy' if p['side'] == 'Sell' else 'Sell',
+                'qty': p['size'],
+                'time_in_force': 'ImmediateOrCancel',
+                'reduce_only': True,
+                'close_on_trigger': True
+            } for p in (r if isinstance(r, list) else [r]) if p['size'] > 0
+        ]
+
+        if len(orders) == 0:
+            return self.logger.error('No position detected.')
+
+        # Submit a market order against each open position for the same qty.
+        return self.place_active_order_bulk(orders)
 
     '''
     Internal methods; signature and request submission.
@@ -1170,7 +1201,7 @@ class HTTP:
     https://bybit-exchange.github.io/docs/inverse/#t-authentication.
     '''
 
-    def _auth(self, method, params):
+    def _auth(self, method, params, recv_window):
         """
         Generates authentication signature per Bybit API specifications.
 
@@ -1190,7 +1221,7 @@ class HTTP:
 
         # Append required parameters.
         params['api_key'] = api_key
-        params['recv_window'] = self.recv_window
+        params['recv_window'] = recv_window
         params['timestamp'] = int(time.time() * 10 ** 3)
 
         # Sort dictionary alphabetically to create querystring.
@@ -1221,47 +1252,15 @@ class HTTP:
 
         """
 
+        # Store original recv_window.
+        recv_window = self.recv_window
+
         # Bug fix: change floating whole numbers to integers to prevent
         # auth signature errors.
-        for i in query.keys():
-            if isinstance(query[i], float) and query[i] == int(query[i]):
-                query[i] = int(query[i])
-
-        # Authenticate if we are using a private endpoint.
-        if auth:
-
-            # Prepare signature.
-            signature = self._auth(
-                method=method,
-                params=query,
-            )
-
-            # Sort the dictionary alphabetically.
-            query = dict(sorted(query.items(), key=lambda x: x))
-
-            # Append the signature to the dictionary.
-            query['sign'] = signature
-
-        # Define parameters and log the request.
         if query is not None:
-            req_params = {k: v for k, v in query.items() if
-                          v is not None}
-        else:
-            req_params = {}
-
-        # Log the request.
-        if self.log_requests:
-            self.logger.info(f'Request -> {method} {path}: {req_params}')
-
-        # Prepare request; use 'params' for GET and 'data' for POST.
-        if method == 'GET':
-            r = self.client.prepare_request(
-                requests.Request(method, path, params=req_params)
-            )
-        else:
-            r = self.client.prepare_request(
-                requests.Request(method, path, data=json.dumps(req_params))
-            )
+            for i in query.keys():
+                if isinstance(query[i], float) and query[i] == int(query[i]):
+                    query[i] = int(query[i])
 
         # Send request and return headers with body. Retry if failed.
         retries_attempted = self.max_retries
@@ -1270,15 +1269,64 @@ class HTTP:
 
             retries_attempted -= 1
             if retries_attempted < 0:
-                raise FailedRequestError('Retries exceeded maximum.')
+                raise FailedRequestError(
+                    message='Bad Request. Retries exceeded maximum.',
+                    status_code=400
+                )
 
             retries_remaining = f'{retries_attempted} retries remain.'
 
+            # Authenticate if we are using a private endpoint.
+            if auth:
+
+                # Prepare signature.
+                signature = self._auth(
+                    method=method,
+                    params=query,
+                    recv_window=recv_window,
+                )
+
+                # Sort the dictionary alphabetically.
+                query = dict(sorted(query.items(), key=lambda x: x))
+
+                # Append the signature to the dictionary.
+                query['sign'] = signature
+
+            # Define parameters and log the request.
+            if query is not None:
+                req_params = {k: v for k, v in query.items() if
+                              v is not None}
+
+            else:
+                req_params = {}
+
+            # Log the request.
+            if self.log_requests:
+                self.logger.info(f'Request -> {method} {path}: {req_params}')
+
+            # Prepare request; use 'params' for GET and 'data' for POST.
+            if method == 'GET':
+                r = self.client.prepare_request(
+                    requests.Request(method, path, params=req_params)
+                )
+            else:
+                r = self.client.prepare_request(
+                    requests.Request(method, path, data=json.dumps(req_params))
+                )
+
+            # Attempt the request.
             try:
                 s = self.client.send(r, timeout=self.timeout)
-            except requests.exceptions.ReadTimeout as e:
+
+            # If requests fires an error, retry.
+            except (
+                requests.exceptions.ReadTimeout,
+                requests.exceptions.SSLError,
+                requests.exceptions.ConnectionError
+            ) as e:
                 if self.force_retry:
                     self.logger.error(f'{e}. {retries_remaining}')
+                    time.sleep(self.retry_delay)
                     continue
                 else:
                     raise e
@@ -1286,23 +1334,59 @@ class HTTP:
             # Convert response to dictionary, or raise if requests error.
             try:
                 s_json = s.json()
+
+            # If we have trouble converting, handle the error and retry.
             except json.decoder.JSONDecodeError as e:
                 if self.force_retry:
                     self.logger.error(f'{e}. {retries_remaining}')
+                    time.sleep(self.retry_delay)
                     continue
                 else:
                     raise FailedRequestError(
-                        f'{s.reason} ({s.status_code})'
+                        message='Conflict. Could not decode JSON.',
+                        status_code=409
                     )
 
             # If Bybit returns an error, raise.
             if s_json['ret_code']:
-                error_msg = f'{s_json["ret_msg"]} ({s_json["ret_code"]})'
-                if self.force_retry:
+
+                # Generate error message.
+                error_msg = f'{s_json["ret_msg"]} (ErrCode: {s_json["ret_code"]})'
+
+                # set default retry delay
+                err_delay = self.retry_delay
+
+                # retry non-fatal whitelisted error requests
+                if s_json['ret_code'] in self.retry_codes:
+
+                    # 10002, recv_window error; add 2.5 seconds and retry.
+                    if s_json['ret_code'] == 10002:
+                        error_msg += '. Added 2.5 seconds to recv_window'
+                        recv_window += 2500
+
+                    # 10006, ratelimit error; wait until rate_limit_reset_ms
+                    # and retry.
+                    elif s_json['ret_code'] == 10006:
+                        self.logger.error(f'{error_msg}. Ratelimited on current request. '
+                                          f'Sleeping, then trying again. Request: {path}')
+
+                        # calculate how long we need to wait.
+                        limit_reset = s_json['rate_limit_reset_ms'] / 1000
+                        reset_str = time.strftime("%X", time.localtime(limit_reset))
+                        err_delay = int(limit_reset) - int(time.time())
+                        error_msg =(f'Ratelimit will reset at {reset_str}. '
+                                    f'Sleeping for {delay} seconds')
+
+                    # Log the error.
                     self.logger.error(f'{error_msg}. {retries_remaining}')
+                    time.sleep(err_delay)
                     continue
+
                 else:
-                    raise InvalidRequestError(error_msg)
+                    raise InvalidRequestError(
+                        message=s_json["ret_msg"],
+                        status_code=s_json["ret_code"]
+                    )
             else:
                 return s_json
 
@@ -1360,6 +1444,9 @@ class WebSocket:
             raise Exception('\'klineV2\' requires a timeframe and ticker, e.g.'
                             ' \'klineV2.5.BTCUSD\'.')
 
+        # set websocket name for logging purposes
+        self.wsName = 'Authenticated' if api_key else 'Non-Authenticated'
+
         # Setup logger.
         logging.basicConfig(
             level=logging_level,
@@ -1367,7 +1454,7 @@ class WebSocket:
             datefmt='%Y-%m-%d %H:%M:%S'
         )
         self.logger = logging.getLogger(__name__)
-        self.logger.info('Initializing WebSocket.')
+        self.logger.info(f'Initializing {self.wsName} WebSocket.')
 
         # Ensure authentication for private topics.
         if any(i in subscriptions for i in [
@@ -1398,12 +1485,8 @@ class WebSocket:
         # Restart on error.
         self.handle_error = restart_on_error
 
-        # Set initial booleans.
-        self.exited = False
-        self.auth = False
-
-        # Initialize dictionary and connect.
-        self.data = {}
+        # Set initial state, initialize dictionary and connnect.
+        self._reset()
         self._connect(self.endpoint)
 
     def fetch(self, topic):
@@ -1421,21 +1504,23 @@ class WebSocket:
 
         # If the topic given isn't in the initial subscribed list.
         if topic not in self.subscriptions:
-            self.logger.error('You aren\'t subscribed to this topic.')
+            self.logger.error(f'You aren\'t subscribed to the {topic} topic.')
             return
 
-        # Pop all trade, execution, or order data on each poll.
+        # Pop all trade or execution data on each poll.
+        # dont pop order or stop_order data as we will lose valuable state
         if topic.startswith((
                 'trade',
-                'execution',
-                'order',
-                'stop_order'
+                'execution'
         )) and not topic.startswith('orderBook'):
             data = self.data[topic].copy()
             self.data[topic] = []
             return data
         else:
-            return self.data[topic]
+            try:
+                return self.data[topic]
+            except KeyError:
+                return []
 
     def ping(self):
         """
@@ -1516,6 +1601,10 @@ class WebSocket:
         if self.api_key and self.api_secret:
             self._auth()
 
+        # Check if subscriptions is a list.
+        if isinstance(self.subscriptions, str):
+            self.subscriptions = [self.subscriptions]
+
         # Subscribe to the requested topics.
         self.ws.send(
             json.dumps({
@@ -1530,11 +1619,11 @@ class WebSocket:
                 self.data[topic] = {}
 
     @staticmethod
-    def _find_index(source, target):
+    def _find_index(source, target, key):
         """
         Find the index in source list of the targeted ID.
         """
-        return next(i for i, j in enumerate(source) if j['id'] == target['id'])
+        return next(i for i, j in enumerate(source) if j[key] == target[key])
 
     def _on_message(self, message):
         """
@@ -1545,34 +1634,32 @@ class WebSocket:
         # Load dict of message.
         msg_json = json.loads(message)
 
-        # If 'success' exists and is True.
-        if 'success' in msg_json and msg_json['success']:
+        # If 'success' exists
+        if 'success' in msg_json:
+            if msg_json['success']:
 
-            # If 'request' exists.
-            if 'request' in msg_json:
+                # If 'request' exists.
+                if 'request' in msg_json:
 
-                # If we get succesful auth, notify user.
-                if msg_json['request']['op'] == 'auth':
-                    self.logger.info('Authorization successful.')
-                    self.auth = True
+                    # If we get succesful auth, notify user.
+                    if msg_json['request']['op'] == 'auth':
+                        self.logger.info('Authorization successful.')
+                        self.auth = True
 
-                # If we get successful subscription, notify user.
-                if msg_json['request']['op'] == 'subscribe':
-                    sub = msg_json['request']['args']
-                    self.logger.info(f'Subscription to {sub} successful.')
+                    # If we get successful subscription, notify user.
+                    if msg_json['request']['op'] == 'subscribe':
+                        sub = msg_json['request']['args']
+                        self.logger.info(f'Subscription to {sub} successful.')
+            else:
+                response = msg_json['ret_msg']
+                if 'unknown topic' in response:
+                    self.logger.error('Couldn\'t subscribe to topic.'
+                                      f' Error: {response}.')
 
-        # If 'success' exists but is False.
-        elif 'success' in msg_json and not msg_json['success']:
-
-            response = msg_json['ret_msg']
-            if 'unknown topic' in response:
-                self.logger.error('Couldn\'t subscribe to topic.'
-                                  f' Error: {response}.')
-
-            # If we get unsuccesful auth, notify user.
-            elif msg_json['request']['op'] == 'auth':
-                self.logger.info('Authorization failed. Please check your '
-                                 'API keys and restart.')
+                # If we get unsuccesful auth, notify user.
+                elif msg_json['request']['op'] == 'auth':
+                    self.logger.info('Authorization failed. Please check your '
+                                     'API keys and restart.')
 
         elif 'topic' in msg_json:
 
@@ -1581,44 +1668,60 @@ class WebSocket:
             # If incoming 'orderbookL2' data.
             if 'orderBookL2' in topic:
 
-                # Record the initial snapshot.
-                if 'snapshot' in msg_json['type']:
-                    self.data[topic] = msg_json['data']
-
+                # Make updates according to delta response.
                 if 'delta' in msg_json['type']:
 
                     # Delete.
                     for entry in msg_json['data']['delete']:
-                        idx = self._find_index(self.data[topic], entry)
+                        idx = self._find_index(self.data[topic], entry, 'id')
                         self.data[topic].pop(idx)
 
                     # Update.
                     for entry in msg_json['data']['update']:
-                        idx = self._find_index(self.data[topic], entry)
+                        idx = self._find_index(self.data[topic], entry, 'id')
                         self.data[topic][idx] = entry
 
                     # Insert.
                     for entry in msg_json['data']['insert']:
                         self.data[topic].append(entry)
 
-            # For incoming 'trade', 'execution', 'order' and 'stop_order'
-            # data.
-            elif any(i in topic for i in ['trade', 'execution', 'order',
-                                          'stop_order']):
+                # Record the initial snapshot.
+                elif 'snapshot' in msg_json['type']:
+                    self.data[topic] = msg_json['data']['order_book']
+
+            # For incoming 'order' and 'stop_order' data.
+            elif any(i in topic for i in ['order', 'stop_order']):
+
+                # record incoming data  
+                for i in msg_json['data']:
+                    try:
+                        # update existing entries
+                        idx = self._find_index(self.data[topic], i, topic+'_id')
+                        self.data[topic][idx] = i
+                    except StopIteration:
+                        # Keep appending or create new list if not already created.
+                        try:
+                            self.data[topic].append(i)
+                        except AttributeError:
+                            self.data[topic] = msg_json['data']
+
+            # For incoming 'trade' and 'execution' data.
+            elif any(i in topic for i in ['trade', 'execution']):
 
                 # Keep appending or create new list if not already created.
                 try:
                     for i in msg_json['data']:
                         self.data[topic].append(i)
                 except AttributeError:
-                    self.data[topic] = [msg_json['data']]
+                    self.data[topic] = msg_json['data']
 
                 # If list is too long, pop the first entry.
                 if len(self.data[topic]) > self.max_length:
                     self.data[topic].pop(0)
 
-            # If incoming 'instrument_info', 'klineV2', or 'wallet' data.
-            elif any(i in topic for i in ['insurance', 'klineV2', 'wallet']):
+            # If incoming 'insurance', 'klineV2', or 'wallet' data.
+            elif any(i in topic for i in ['insurance', 'klineV2', 'wallet',
+                                          'candle']):
 
                 # Record incoming data.
                 self.data[topic] = msg_json['data'][0]
@@ -1626,21 +1729,34 @@ class WebSocket:
             # If incoming 'instrument_info' data.
             elif 'instrument_info' in topic:
 
-                # Record the initial snapshot.
-                if 'snapshot' in msg_json['type']:
-                    self.data[topic] = msg_json['data']
-
                 # Make updates according to delta response.
-                elif 'delta' in msg_json['type']:
+                if 'delta' in msg_json['type']:
                     for i in msg_json['data']['update'][0]:
                         self.data[topic][i] = msg_json['data']['update'][0][i]
 
+                # Record the initial snapshot.
+                elif 'snapshot' in msg_json['type']:
+                    self.data[topic] = msg_json['data']
+
             # If incoming 'position' data.
             elif 'position' in topic:
 
                 # Record incoming position data.
-                data = msg_json['data'][0]
-                self.data[topic][msg_json['data'][0]['symbol']] = data
+                for p in msg_json['data']:
+
+                    # linear (USDT) positions have Buy|Sell side and
+                    # updates contain all USDT positions.
+                    # For linear tickers...
+                    if p['symbol'].endswith('USDT'):
+                        try:
+                            self.data[topic][p['symbol']][p['side']] = p
+                        # if side key hasn't been created yet...
+                        except KeyError:
+                            self.data[topic][p['symbol']] = {p['side']: p}
+
+                    # For non-linear tickers...
+                    else:
+                        self.data[topic][p['symbol']] = p
 
     def _on_error(self, error):
         """
@@ -1648,21 +1764,30 @@ class WebSocket:
         """
 
         if not self.exited:
-            self.logger.error(f'WebSocket encountered error: {error}.')
+            self.logger.error(f'WebSocket {self.wsName} encountered error: {error}.')
             self.exit()
 
         # Reconnect.
         if self.handle_error:
+            self._reset()
             self._connect(self.endpoint)
 
     def _on_open(self):
         """
         Log WS open.
         """
-        self.logger.debug('WebSocket opened.')
+        self.logger.debug(f'WebSocket {self.wsName} opened.')
 
     def _on_close(self):
         """
         Log WS close.
         """
-        self.logger.info('WebSocket closed.')
+        self.logger.info(f'WebSocket {self.wsName} closed.')
+
+    def _reset(self):
+        """
+        Set state booleans and initialize dictionary.
+        """
+        self.exited = False
+        self.auth = False
+        self.data = {}
